#include "fahren.h"
#include "nibobee.h"

// Antriebsmotoren regeln

// Zur Dokumentation der Anwendungs-Schnittstelle siehe servo.h


// Konfigurations-Parameter für die Motor-Steuerung
// Fließkomma-Zahlen müssen als Bruch dargestellt werden, damit
// der Compiler keine teure Fließkomma-Arithmetik nutzt.
// Beispiel: Statt 0.5 schreibt man 1/2.

#define BESCHLEUNIGUNG      2        // Beschleunigung und Bremsung pro 100ms Intervall und Fehler, bei zuviel ruckelt das Fahrzeug 
                                     // beim Beschleunigen oder Bremsen.

#define LENKUNG_P           8        // Lenkung proportional pro Fehler bezogen auf die gesamte Strecke.

#define LENKUNG_I           2        // Lenkung integral pro 100ms Intervall und Fehler.

#define START_PWM           300      // PWM Wert, bei dem die Motoren aus dem Stand heraus sicher zu drehen beginnen.

#define MAX_BLOCKADE        20       // Maximale Anzahl von 100ms Intervallen für blockierte Motoren.

#define SPERRZEIT           2000     // Abkühlzeit in ms nach einer Blockade.

#define RUHEZEIT            100      // Wartezeit nach jeder Teilstrecke, um Schwingungen zu beruhigen

#define BREMSKRAFT          10       // Zur Abschätzung des Bremsweges. Wenn zu groß, bremst der Roboter zu abrupt.


// Werte, die erreicht werden sollen
static int32_t soll_distanz; // Distanz in Odometer Impulse
static int16_t soll_lenkung; // Lenkung in Promille 0-1000, 500=geradeaus
static int16_t soll_tempo;   // Odometer Impulse pro Sekunde, 0=nicht fahren

// Aktuelle Geschwindigkeit (des schnelleren Rades), wird alle 100ms aktualisiert
static int16_t tempo;

// Bis zu dieser Zeit muss vor dem nächsten Fahrmanöver gewartet werden.
static uint32_t sperrzeit=0;

// Anzahl von 100ms Intervallen, die einer oder beide Motoren gerade blockiert sind
static uint8_t  blockade_zaehler;

// Aktueller Status der Motorsteuerung
static uint8_t status=0;


void fahren_regeln() {
    static uint32_t letzter_zeitpunkt; // Zeitpunkt des vorherigen Regel-Intervalls
    static uint32_t letzter_odo_l;     // Position links beim vorherigen Regel-Intervall
    static uint32_t letzter_odo_r;     // Position rechts beim vorherigen Regel-Intervall
    static int16_t leistung;           // Aktuelle Leistung (pro Motor, 0-1023)
    static int16_t lenkung_i;          // Aktuelle Lenkung in Promille, nur der integrale Anteil (0-1000)
    
    if (status==STARTET) { 
      
        // aber erst wenn die Sperrzeit abgelaufen ist
        if (system_time()>sperrzeit) { 
            // Variablen für erste Intervall-Messung vorbereiten   
            blockade_zaehler=0;
	        reset_odometer();
            letzter_odo_l=0;
            letzter_odo_r=0;
            letzter_zeitpunkt=system_time();
            // Motoren starten
            if (soll_lenkung>0) 
                PWM_L=START_PWM;
            else
                PWM_L=0;
            if (soll_lenkung<1000) 
                PWM_R=START_PWM;  
            else
                PWM_R=0;
            // Aktuelle Lenkung und Leistung setzen, wird während der Fahrt geregelt
	       lenkung_i=soll_lenkung;
            leistung=PWM_L+PWM_R;
            // Zur nächsten Phase übergehen
            status=FAEHRT;
        }
    } 
    
    else if (status==FAEHRT) { 
        
        // Wenn das Ziel gerade erreicht wurde: motoren stoppen, nicht mehr weiter regeln
        if (odometer_left()+odometer_right() >= soll_distanz<<1) {
            PWM_R=0;
            PWM_L=0;
	        sperrzeit=system_time()+RUHEZEIT; 
            status=ZIEL_ERREICHT;
            return;
        }      
            
        // In regelmäßigen Intervallen Geschwindigkeit messen und Motoren regeln
        uint16_t intervall_zeit=system_time()-letzter_zeitpunkt;
        if (intervall_zeit>100) {
            
            // Ermittle die gefahrene Strecke im gerade verstrichenen Intervall
            int32_t intervall_odo_l=odometer_left()-letzter_odo_l;
            int32_t intervall_odo_r=odometer_right()-letzter_odo_r;
                       
            // Variablen für die nächste intervall-Messung vorbereiten
            letzter_zeitpunkt=system_time();
            letzter_odo_l=odometer_left();
            letzter_odo_r=odometer_right();
	      
            // Wenn ein Motor zu lange blockiert wurde: motoren stoppen, nicht mehr weiter regeln
            if ((intervall_odo_l==0 && PWM_L>100) || (intervall_odo_r==0 && PWM_R>100)) {
                blockade_zaehler++;
                if (blockade_zaehler>MAX_BLOCKADE) {
                    // Stoppen
                    PWM_L=0;
                    PWM_R=0;
                    // Die nächste Fahrt darf erst nach einer Pause beginnen
                    // damit die Motoren und Transistoren abkühlen können.
                    sperrzeit=system_time()+SPERRZEIT; 
                    status=MOTOR_BLOCKIERT;
                    return;
                }
            }
            else {
                blockade_zaehler=0;
            }                
                       
            // Berechne aktuelle Geschwindigkeit. Das schnellere Rad zählt.
            int32_t intervall_odo;
            if (intervall_odo_l>intervall_odo_r) 
                intervall_odo=intervall_odo_l;
            else
                intervall_odo=intervall_odo_r; 
            tempo=intervall_odo*(1000/intervall_zeit);  
	           
	        // Geschwindigkeit so begrenzen, daß der Anhalteweg das Ziel
            // nicht überschreitet. Ich verwende eine Annäherungsschleife,
            // weil ich sonst Wurzel ziehen müsste, was zu viel Rechenzeit
            // in Anpruch nehmen würde.
	        int32_t distanz=(odometer_left()+odometer_right())>>1;
            int32_t rest_distanz=soll_distanz-distanz;
            while(soll_tempo>40) { // Bei sehr niedrigem Temp ist sanftes Bremsen nicht nötig
                // Anhalteweg ist reaktionsweg + bremsweg
                int16_t anhalteweg=intervall_odo+((soll_tempo>>2)*(soll_tempo>>2) / BREMSKRAFT);  
                if (anhalteweg>rest_distanz) {
                    soll_tempo-=soll_tempo>>2;
                }
                else break;
            } 
             
            // Korrekturwert der Leistung berechnen 
            int16_t korrektur=(soll_tempo-tempo)*BESCHLEUNIGUNG;

            // Leistung korrigieren
            leistung+=korrektur;
               
            // Lenkung nur regeln, wenn wir nicht sowieso bis zum Anschlag lenken 
            // und außerdem erst nach einigen Millimetern Fahrt regeln.   
            int16_t lenkung; 
            if (soll_lenkung>0 && soll_lenkung<1000 && odometer_left()>2 && odometer_right()>2) {
                // Fehler berechnen
                int16_t fehler=(int32_t) soll_lenkung*odometer_right()/(1000-soll_lenkung)-odometer_left();
                
                // Integral- und proportional Anteil berechnen
                lenkung_i+=fehler*LENKUNG_I;
                int16_t lenkung_p=fehler*LENKUNG_P;
                lenkung=lenkung_i+lenkung_p;
                
                // Lenkung auf gültigen Bereich beschränken
                if (lenkung>1000)
                    lenkung=1000;
                else if (lenkung<0)
                    lenkung=0; 
            }
            else {
                // Lenkung nicht regeln
                lenkung=soll_lenkung;
            } 
	    
           
            // Leistung so begrenzen, daß die PWM nicht übersteuert wird.
            // Die Maximale Leistung liegt zwischen 1023 und 2047, je nach Lenkung.
            if (leistung>0) {
                int16_t max_leistung;
                if (lenkung>500)
                    max_leistung=1023000/lenkung;
                else if (lenkung<500)
                    max_leistung=1023000/(1000-lenkung);
                else
                    max_leistung=2047;
                if (leistung>max_leistung)
                    leistung=max_leistung;
            }
            else {
                leistung=0; // Für den Fall, daß sie negativ war
            }
            
            // Motoren ansteuern
            PWM_L=(int32_t)leistung*lenkung/1000;
	        PWM_R=(int32_t)leistung*(1000-lenkung)/1000;
        }       
    }
}



void fahre(int32_t distanz, int16_t tempo, int16_t lenkung) {   
  
    // Kehre das Vorzeichen um, wenn das Tempo negativ ist
    if (tempo<0) {
        distanz*=-1;
        tempo*=-1;
    }

    // Stelle Dreh-Richtung der Motoren ein
    if (lenkung>=0) {
        // Vorwärts oder Rückwärts fahren
        set_DIR_R(distanz<0);
        set_DIR_L(distanz>0);
    }
    else {
        // Auf der Stelle drehen (also beide Räder entgegengesetzt drehen)
        set_DIR_R(distanz>0);
        set_DIR_L(distanz>0);
    }

    // Entferne ggf. negative Vorzeichen von Distanz und Lenkung
    if (lenkung<0) lenkung*=-1;
    if (distanz<0) distanz*=-1;
    
    // Setzte die Soll-Werte für die Fahrt
    soll_lenkung=lenkung;
    soll_tempo=tempo;
    soll_distanz=distanz; 
    
    // Beginne die Fahrt
    status=STARTET;
}



void stop() {
    PWM_L=0;
    PWM_R=0;
    status=GESTOPPT;
}



uint16_t get_tempo() {
    return tempo;
}



uint16_t get_strecke() {
    return (odometer_left()+odometer_right())>>1;
}



uint8_t fahrstatus() {
    return status;
}
