// Diese Datei enthält Funktionen zum regeln der Motoren, um Strecken
// mit bestimmter Geschwindigkeit und Kurven in bestimmtem Radius
// zu fahren.

// Folgende Funktionen sind implementiert:
//     - exaktes geradeaus Fahren, sowie Kurven mit bestimmtem Radius und Winkel.
//     - sanftes Anfahren
//     - rechtzeitiges sanftes Abbremsen vor dem Ziel
//     - fortlaufende Erfassung von Distanz und Geschwindigkeit
//     - Erkennung von blockierten Motoren, mit Abkühl-Timer

// Anwendungsbeispiel:

//     vorwaerts(100*CM,SCHNELL);
//     while (fahrstatus() < ZIEL_ERREICHT) {
//         fahren_regeln();
//     }

// Bekannte Mängel:
//     - Jeder Fahrbefehl endet mit einer sanften Bremsung. Es ist nicht möglich,
//       zwei aufeinanderfolgende geradeaus-Teilstrecken ohne Zwischenstopp zu durchfahren.
//     - Die Odometrie Sensoren liefern ungefähr alle 3mm bzw 3 Grad einen Impuls und
//       die Motoren können nur im 120 Grad Raster stoppen, was die Genauigkeit entsprechend 
//       beschränkt. Tendenziell stoppt der Roboter eher etwas zu spät, als zu früh.

// In der fahren.c befinden sich weitere Konfigurationsparameter.

// Autor: Stefan Frings <stefan@stefanfrings.de>
// Datum: 01.12.2010


#ifndef _FAHREN_H_
#define _FAHREN_H_

#include "nibobee.h"
#include <stdint.h>


// Odometrie Impulse pro Zentimeter, hängt vom Rad-Durchmesser ab.
#define CM (int32_t) 7/2

// Odometrie Impulse pro Grad bei Drehungen, hängt von Rad-Durchmesser und Achs-Breite ab.
#define GRAD (int32_t) 1/3


// Gängige Geschwindigkeiten 
#define SCHNECKE     6*CM   // weniger lässt sich nicht regeln
#define LANGSAM      20*CM
#define MITTEL       40*CM
#define SCHNELL      100*CM  
#define SUPERSCHNELL 250*CM // schaffen die Motoren nur ohne Last


// Mögliche Stati der Motorsteuerung
#define GESTOPPT        0    // gestoppt
#define STARTET         1    // gestartet, dauert ggf. bis zum Ende der Sperrzeit
#define FAEHRT          2    // fährt
#define ZIEL_ERREICHT   3    // Ziel erreicht
#define MOTOR_BLOCKIERT 4    // Ein Motor ist blockiert
// Stati 4 und 5 bleiben bis zum Nächsten Fahrbefehl oder Stopp bestehen.


// Fragt den aktuellen Status der Motorsteuerung ab.
// Mögliche Rückgabewerte: siehe oben.
uint8_t fahrstatus();



// Diese Funktion soll so oft wie möglich immer wieder aufgerufen werden.
// Sie regelt die Ansteuerung der Motoren.
void fahren_regeln();



// Fahre eine bestimmte Strecke mit einer bestimmten Geschwindigkeit. Zum
// rückwärts Fahren muss entweder die Distanz oder das Tempo negativ sein.
// Die Funktion kehrt sofort zurück, die Fahrt beginnt tatsächlich erst 
// beim folgenden Aufruf der fahren_regeln() Funktion. 
// Parameter:
//     distanz in Odometer Takten
//     tempo in Odometer Takten pro Sekunde (20-250)
//     lenkung  (0=ganz link, 500=geradeaus, 1000=ganz rechts, -500=auf der Stelle drehen)
void fahre(int32_t distanz, int16_t tempo, int16_t lenkung);


// Die folgenden Makros können alternativ zur fahre(...) Funktion verwendet werden.

// Geradeaus fahren 
#define vorwaerts(distanz,tempo)   fahre(distanz,tempo,500)
#define rueckwaerts(distanz,tempo) fahre(distanz*-1,tempo,500)

// Rotiere um die eigene Achse (positiver Winkel=rechts herum)
#define rotiere(winkel,tempo) fahre(winkel,tempo,-500)

// Rechtskurve zu einem bestimmten Winkel (lenkung muss 501-1000 sein, winkel>0)
// Zum Rückwärts fahren, Winkel oder Tempo negativ wählen
#define rechtskurve(winkel,tempo,lenkung) fahre((int32_t) winkel*500/(lenkung-500),tempo,lenkung)

// Linkskurve zu einem bestimmten Winkel (lenkung muss 0-499 sein, winkel>0)
// Zum Rückwärts fahren, Winkel oder Tempo negativ wählen
#define linkskurve(winkel,tempo,lenkung) fahre((int32_t) winkel*500/(500-lenkung),tempo,lenkung)



// Stoppt die Fahrt durch Vollbremsung.
void stop();



// Fragt die aktuelle Geschwindigkeit des schnelleren Rades ab.
// Der Wert wird alle 100ms durch fahren_regeln() aktualisiert.
uint16_t get_tempo();



// Erfragt die gefahrene Strecke (Odometrie-Impulse), die seit beginn der 
// aktuellen Fahrt zurück gelegt wurde. Der Wert wird fortlaufend durch 
// fahren_regeln() aktualisiert.
uint16_t get_strecke();


#endif // _FAHREN_H_
