#include "nibobee.h"

// Zur Dokumentation der Anwendungs-Schnittstelle siehe servo.h

#include <avr/interrupt.h>

#if USE_SERIAL
    #include <stdio.h>
    #include <util/setbaud.h>
#endif

// Alias zur Kompatibilität mit den neureren AVR Mikrocontrollern
// die zwei serielle Ports haben
#ifdef UDR0
    #define UDR UDR0
    #define UBRRL UBRR0L
    #define UBRRH UBRR0H
    #define UCSRA UCSR0A
    #define UCSRB UCSR0B
    #define UCSRC UCSR0C
    #define RXEN  RXEN0
    #define TXEN TXEN0
    #define UCSZ0 UCSZ00
    #define UCSZ1 UCSZ01
    #define U2X U2X0
    #define RXC RXC0
    #define UDRE UDRE0
#endif

// Alias zur Kompatibilität mit neueren AVR Mikrocontrollern
// die separate Interrupts für Timer0 COMPA und COMPB haben
#ifdef TIMER0_COMPA_vect
    #define TIMER0_COMP_vect TIMER0_COMPA_vect
#endif


// Werte vom ADC
// Index 0-7 sind Kanal 0-7 mit PB4=low
// Index 8-15 sind Kanäle 0-7 mit PB4=high
static volatile uint16_t adc_value[16];


// Odometrie Zähler, 40 Impulse pro Rad-Umdrehung
static volatile uint32_t odometer_l;
static volatile uint32_t odometer_r;


// System Zeitmesser in Millisekunden
static volatile uint32_t systime;


// Der folgende Code bindet den seriellen Port ans stdin und stdout
#if USE_SERIAL
    static int serial_write(char, FILE *);
    static int serial_read(FILE *);
    static FILE serialPort = FDEV_SETUP_STREAM(serial_write, serial_read, _FDEV_SETUP_RW);
    
    // Sende ein Zeichen
    static int serial_write(char c, FILE *f) {
        #if TERMINAL_MODE
            if (c=='\n') {
               loop_until_bit_is_set(UCSRA, UDRE);
               UDR='\r';
           }
        #endif /* TERMINAL_MODE */
        loop_until_bit_is_set(UCSRA, UDRE);
        UDR = c;
        return 0;
    }


    // Lese ein Zeichen
    static int serial_read(FILE *f) {
        while (readBit(UCSRA, RXC)==0);
        char c=UDR;
        #if SERIAL_ECHO
            loop_until_bit_is_set(UCSRA, UDRE);
           UDR = c;
        #endif
        #if TERMINAL_MODE
            if (c=='\r') {
                c='\n';
                #ifdef SERIAL_ECHO
                    loop_until_bit_is_set(UCSRA, UDRE);
                    UDR = c;
                #endif 
            }
        #endif
        return c;
    }


    // Initialisiere den seriellen Port
    void initserial(void) {
        // set baudrate
        UBRRH = UBRRH_VALUE;
        UBRRL = UBRRL_VALUE;
        #if USE_2X
            UCSRA |= (1 << U2X);
        #else
            UCSRA &= ~(1 << U2X);
        #endif
        // Aktiviere Empfänger und Sender
        UCSRB = (1<<RXEN) | (1<<TXEN);
        // framing format 8N1
        #ifdef URSEL
            UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0);
        #else
            UCSRC = (1<<UCSZ1) | (1<<UCSZ0);
        #endif
        // Binde stdout and stdin an den seriellen Port
        stdout = &serialPort;
        stdin = &serialPort;
    }
#endif /* USE_SERIAL */


// Unterbrechung vom ADC
ISR(ADC_vect) {
    uint16_t value = ADC;
    uint8_t channel = ADMUX & 7;
    uint8_t pb4 = readBit(PORTB,4);
    adc_value[channel+(pb4<<3)]=value;
    // Zähle von 0-7 und schalte dann PB4 bei Überlauf um.
    if (++channel > 7) {
        PORTB ^= (1<<4);
        channel=0;
    }
    // Schalte zum nächsten Kanal
    ADMUX=(ADMUX & ~7) | channel;
    // Starte nächste Messung
    ADCSRA |= (1<<ADSC); 
}



void set_AREF(uint8_t input) {
    ADMUX=(ADMUX & ~((1<<REFS0) | (1<<REFS1))) | input;
}



// Unterbrechung vom linken Odometrie Sensor
ISR(INT0_vect) {
    odometer_l++;
}



// Unterbrechung vom rechten Odometrie Sensor
ISR(INT1_vect) {
    odometer_r++;
}



void reset_odometer() {
    odometer_l = 0;
    odometer_r = 0;
}


// Timer 0 Unterbrechung (75khz).
// Inkrementiert system_time bei jedem 75ten Aufruf.
ISR(TIMER0_COMP_vect) {
    static uint8_t prescaler;
    if (--prescaler==0) {
        prescaler=75;
        systime++;
    }    
}


void batterie_status() {
    set_AREF(REF_256);
    for (uint8_t i=0; i<6; i++) {
        sleep(50);
        set_LED0(analog(VBAT)>(410*2));
        set_LED1(analog(VBAT)>(430*2));
        set_LED2(analog(VBAT)>(450*2));
        set_LED3(analog(VBAT)>(470*2));
        sleep(50);
        set_LED0(0);
        set_LED1(0);
        set_LED2(0);
        set_LED3(0);
    }
    set_AREF(REF_VCC);
    sleep(50);
}



// Installiert die nibobee_init() Prozedur in den Start-Bereich 5,
// so dass sie automatisch vor main() ausgeführt wird.
void nibobee_init() __attribute__ ((naked))  __attribute__ ((section (".init5")));


// Initialisiert die Hardware
void nibobee_init() {
    
    // Setzt die Ausgabe PIN's und Pull-up Wiederstände
    DDRB |= 1+2+4+8+16; // LED0-3 and LINE_EN are outputs
    DDRD |= 16+32+64+128; // PWM_R, PWM_L, DIR_L and DIR_R are outputs
    PORTC |= 16+32+64+128; // Enable Pull-ups for sensor switches

    // Fahre vorwärts
    set_DIR_L(1); // Linker Motor gegen den Uhrzeigersinn
    set_DIR_R(0); // Rechter Motor im Uhrzeigersinn
    
    // Initialisiere den seriellen Port
    #if USE_SERIAL
        initserial();
    #endif /* USE_SERIAL */
    
    // Initialisiere die PWM Erzeugung (Timer 1).
    // Modus 2 = Phasenrichtige 10 bit PWM,
    // Kanal A+B werden bei Erreichen des Soll beim hochzählen gesetzt und
    // bei Erreichen des Soll beim herunterzählen gelöscht.
    TCCR1A = (1<<COM1A1) | (1<<COM1A0) | (1<<COM1B1) | (1<<COM1B0) | (1<<WGM11) | (1<<WGM10);
    // Use I/O clock without prescaling
    TCCR1B = (1<<CS10);

    
    // Initialisiere die Ododmetrie Zähler (Software Zähler).
    // Löse Interrupts bei jedem Wechseland INT0 und INT1 aus.
    #ifdef EICRA
        // ATmega644
        EICRA |= (1<<ISC10) | (1<<ISC00);
        EIMSK |= (1<<INT1) |  (1<<INT0);
    #else
        // ATmega16
        MCUCR |= (1<<ISC10) | (1<<ISC00);
        GICR |= (1<<INT1) |  (1<<INT0);
    #endif
    odometer_l = 0;
    odometer_r = 0;
    
    
    // Initialisiere den ADC
    // Aktiviere ADC Konvertierung im Einzelschritt Modus mit 117khz
    // Taktfrequenz (F_CPU/128) , was etwa 2000 Messungen pro Sekunde für
    // alle 8 Kanäle ergibt.
    ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0) | (1<<ADIE) | (1<<ADSC); 
    ADMUX = REF_VCC;

   
    // Initialisiere den System Zeitmesser (Timer 0).
    // Prescaler 8 und Counter 25 -> 75khz
    // Modus: Setze Zähler bei Erreichen des Soll zurück, keinen Vorteiler
    #ifdef TIMSK0
        // Für neue AVR Modelle
        OCR0A = 25;
        TCCR0A = (1<<WGM01);
        TCCR0B = (1<<CS01);
        TIMSK0 |= (1<<OCIE0A);
    #else
        // ATmega16
        OCR0 = 25;
        TCCR0 = (1<<WGM01) | (1<<CS01);
        TIMSK |= (1<<OCIE0);
    #endif

    // TWI Interface ungefähr 100kbit/section
    TWBR=75;

    // Erlaube Unterbrechungen
    sei();
}


uint8_t i2c_communicate(uint8_t slave_address, void* send_data, uint8_t send_bytes, void* rcv_data, uint8_t rcv_bytes) {
    uint8_t rcv_count=0;
            
    // Adresse ein Bit nach links verschieben, um Platz für das r/w Flag zu schaffen
    slave_address=slave_address<<1;
           
    if (send_bytes>0) {
        // Sende Start 
        TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWSTA);
        while (!(TWCR & (1<<TWINT)));
        uint8_t status=TWSR & 0xf8;
        if (status != 0x08 && status != 0x10) goto error;
    
        // Sende Adresse (write mode)
        TWDR=slave_address;
        TWCR=(1<<TWINT) | (1<<TWEN);
        while (!(TWCR & (1<<TWINT)));
        if ((TWSR & 0xf8) != 0x18) goto error;
        
        // Sende Daten
        while (send_bytes>0) {
            TWDR=*((uint8_t*)send_data);
            TWCR=(1<<TWINT) | (1<<TWEN);
            while (!(TWCR & (1<<TWINT)));
            if ((TWSR & 0xf8) != 0x28) goto error;
            send_data++;
            send_bytes--;
        }
    }
    
    if (rcv_bytes>0) { 
        // Sende START
        TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWSTA);
        while (!(TWCR & (1<<TWINT)));
        uint8_t status=TWSR & 0xf8;
        if (status != 0x08 && status != 0x10) goto error;
        
        // Sende Adresse (read mode)
        TWDR=slave_address + 1;
        TWCR=(1<<TWINT) | (1<<TWEN);
        while (!(TWCR & (1<<TWINT)));
        if ((TWSR & 0xf8) != 0x40) goto error;
    
        // Empfange Daten
        while (rcv_bytes>0) {    
            if (rcv_bytes==1) {
                // das letzte Byte nicht mit ACK quittieren
                TWCR=(1<<TWINT) | (1<<TWEN); 
            } 
            else {   
                // alle anderen Bytes mit ACK quittieren
                TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWEA);
            }
            while (!(TWCR & (1<<TWINT)));
            uint8_t status=TWSR & 0xf8;
            if (status!=0x50 && status != 0x58) goto error;
            *((uint8_t*)rcv_data)=TWDR;
            rcv_data++;
            rcv_bytes--;
            rcv_count++;
        }

    }
    
    // Sende STOP
    TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
    return rcv_count;

    error:
    // Sende STOP
    TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
    return 0;
}


uint16_t analog(uint8_t channel) {
    cli();
    uint16_t i=adc_value[channel];
    sei();
    return i;
}


uint32_t odometer_left() {
    cli();
    uint32_t l=odometer_l;
    sei();
    return l;
}


uint32_t odometer_right() {
    cli();
    uint32_t r=odometer_r;
    sei();
    return r;
}


uint32_t system_time() {
    cli();
    uint32_t l=systime;
    sei();
    return l;
}


void sleep(uint32_t milliseconds) {
    cli();
    uint32_t start=systime;
    uint32_t time=systime;
    sei();
    while (time-start < milliseconds) {
        cli();
        time=systime;
        sei();
    };
}
