// Diese Datei enthält Makros, Definitionen und Funktionen, um die Hardware
// des NIBObee anzusprechen.

// Autor: Stefan Frings <stefan@stefanfrings.de>
// Datum: 02.08.2012

#ifndef _NIBOBEE_H_
#define _NIBOBEE_H_

#include <stdint.h>
#include <avr/io.h>



//===========================================================================
// Zugriff auf einzelne Port Bits
//===========================================================================
// Diese Makros sind primär dazu gedacht, auf einzelne I/O Pins zuzugreifen,
// sie können aber auch auf Register und Variablem angewendet werden. Diese
// Maros werden vom Compiler ziemlich gut optimiert, also verwende sie,
// wo immer es den Programmcode beser lesbar macht. Beispiele:
// writeBit(PORTB,3,1);
// uint8_t portC7=readBit(PINC,7);

// Setze ein Bit auf high, wenn value>0 ist, andernfalls setze das Bit auf low
#define writeBit(port,bit,value)   { if ((value)>0) (port) |= (1<<bit); else (port) &= ~(1<<bit); }

// Lese ein Bit. Das Ergebnis ist immer 0 oder 1.
#define readBit(port,bit)          (((port) >> (bit)) & 1)




//===========================================================================
// Serieller port
//===========================================================================
// Um den seriellen Port als stdin und stdout zu verwenden, müssen im
// Makefile vier Definitionen gesetzt werden:
//
// -DUSE_SERIAL=1 -DBAUD=115200 -DTERMINAL_MODE=1 -DSERIAL_ECHO=1
//
// Zum Deaktivieren genügt eine Definition:
//
// -DUSE_SERIAL=0
//
// Der Terminal Modus aktiviert Unterstützung für Programme, die 
// Zeilenumbrüche im DOS Format erwarten und senden. Alle Funktionen 
// aus stdio.h können den seriellen Port nutzen, falls aktiviert.
//
// Ein Zeichen empfangen: char c=getchar();
//
// Ein Zeichen senden: putchar(c);
//
// Eine Zeile empfangen: char buffer[100];
//                       fgets(buffer,sizeof(buffer),stdin);
//
// Eine Zeile senden: puts("Hallo\n");
//
// Der serielle Port ist ungepuffert. Sowohl Eingabe als auch Ausgabe
// blockieren, bis ein Zeichen verfügbar ist bzw. gesendet wurde.

// Dieses Makro kann verwendet werden, um zu testen, ob der serielle
// Port ein zeichen empfangen hat, bevor man es ausliest (z.B. mit getch()).
#define RECEIVED readBit(UCSRA, RXC)



//===========================================================================
// Analoge Eingänge
//===========================================================================
// Die analogen Eingänge werden im Hintergrund durch eine Interrupt-Routine
// ausgelesen und in ein Array eingetragen. Die Funktion analog() kann 
// benutzt werden, um das Array zu lesen. Jeder Kanal wird zweimal gemessen:
// einmal während PB4 high ist, und einmal während PB4 low ist. PB4 schaltet 
// das Licht vom Linien-Sensor an und aus.
//
// Linien Sensoren sollten mit VCC als Referenz gelesen werden (was der 
// Vorgabe entspricht). Der Wert liegt typischerweise unter 600 bei dunklen 
// Flächen und über 600 bei hellen Flächen.
//
// Um Sinnvolle Werte vom Batteriestatus zu erhalten, muss man allerdings
// 2.56V als Referenz benutzen.

// Analoge Eingänge, PB4=low (Liniensensor Licht an)
#define AN0 0
#define AN1 1
#define AN2 2
#define AN3 3
#define VBAT 4
#define LINE_L 5
#define LINE_C 6
#define LINE_R 7

// Analoge Eingänge, PB4=high (Liniensensor Licht aus)
#define _AN0 8
#define _AN1 9
#define _AN2 10
#define _AN3 11
#define _VBAT 12
#define _LINE_L 13
#define _LINE_C 14
#define _LINE_R 15

// Werte vom ADC auslesen
uint16_t analog(uint8_t channel);


// Analoge Referenz Werte (VCC oder 2.56V)
#define REF_VCC (1<<REFS0)
#define REF_256 ((1<<REFS0) | (1<<REFS1))

// Ändere die Referenz-Spannung des ADC, siehe oben für gültige input Werte.
// Nach Änderung muß man etwas warten (z.B. 100ms), um C16 zu laden.
void set_AREF(uint8_t input);

// Zeige den Batterie-Zustand durch mehrmals blinkende LED's an. 
// DIe Funktion blockiert den Programmablauf, bis sie fertig ist.
// LED0 = >4.1V
// LED1 = >4.3V
// LED2 = >4.5V
// LED3 = >4.7V
void batterie_status();



//===========================================================================
// Sensor Taster
//===========================================================================
// Diese Makros liefern 1 zurück, wenn ein Taster gedrückt wurde

#define SENS_SW1 (!readBit(PINC,4))
#define SENS_SW2 (!readBit(PINC,5))
#define SENS_SW3 (!readBit(PINC,7))
#define SENS_SW4 (!readBit(PINC,6))

// Alternativen zu den obeigen Makros
#define SENS_L1 (!readBit(PINC,4))
#define SENS_L2 (!readBit(PINC,5))
#define SENS_R1 (!readBit(PINC,6))
#define SENS_R2 (!readBit(PINC,7))

// Kombinationen von 2 bzw. 4 Tastern
#define SENS_LEFT  ((PINC & ((1<<4)+(1<<5))) != ((1<<4)+(1<<5)))
#define SENS_RIGHT ((PINC & ((1<<6)+(1<<7))) != ((1<<6)+(1<<7)))
#define SENS_ANY   ((PINC & ((1<<4)+(1<<5)+(1<<6)+(1<<7))) != ((1<<4)+(1<<5)+(1<<6)+(1<<7)))



//===========================================================================
// Status LED's
//===========================================================================
// Die Status LED's sind frei programmierbar.

// Die LED geht an, wenn der Wert nicht 0 ist
#define set_LED0(value) writeBit(PORTB,0,value)
#define set_LED1(value) writeBit(PORTB,1,value)
#define set_LED2(value) writeBit(PORTB,2,value)
#define set_LED3(value) writeBit(PORTB,3,value)

// Liefert 1 zurück, wenn die LED leuchtet
#define LED0 readBit(PINB,0)
#define LED1 readBit(PINB,1)
#define LED2 readBit(PINB,2)
#define LED3 readBit(PINB,3)



//===========================================================================
// Motor Steuerung
//===========================================================================
// Die Dreh-Richtung wird für jede Seite durch einen eigenen Port-PIN
// festgelegt. Die Leistung wird mit 10-Bit PWM (Timer 1) gesteuert.
// Aus dem Stand heraus beginnt der Motor ab Wert 250, sich zu drehen.

// Motor Richtung festlegen
// 0=Uhrzeigersinn, 1=gegen den Urzeigersinn 
#define set_DIR_L(value) writeBit(PORTD,6,value)
#define set_DIR_R(value) writeBit(PORTD,7,value)

// Motor Richtung abfragen
#define DIR_L readBit(PIND,6)
#define DIR_R readBit(PIND,7)

// Motor Leistung festlegen oder abfragen
// 0=Stop, 1023=maximal
#define PWM_L OCR1A
#define PWM_R OCR1B



//===========================================================================
// Odometrie Sensoren
//===========================================================================
// Die Odometrie Sensoren erzeugen 40 Unterbrechungen pro Rad-Umdrehung,
// welche gezählt werden.

// Odometer Sensoren, aktuellen Zustand abfragen
#define ODO_L readBit(PIND,2)
#define ODO_R readBit(PIND,3)

// Setze Odometrie Zähler zurück
void reset_odometer();

// Odometrie Zähler auslesen
uint32_t odometer_left();
uint32_t odometer_right();



//===========================================================================
// System Zeitmesser
//===========================================================================
// Der System Zeitmesser zählt Unterbrechungen von Timer 0. Er zählt die 
// Anzahl von Millisekunden seit dem letzten Reset.

// System Zeitmesser in Millisekunden abfragen
uint32_t system_time();



//===========================================================================
// Warteschleife
//===========================================================================

// Wartet eine bestimmte Anzahl von Millisekunden.
// Diese Funktion blockiert den Programmablauf bis zum Ablauf der Zeit.
void sleep(uint32_t milliseconds);



//===========================================================================
// I2C Bus
//===========================================================================

// Wenn der I2C Bus nicht verwendet wird, stehen dessen Pins als reguläre
// I/O Pins zur Verfügung. Pull-Up Widerstände nicht vergessen!

// Sendet beliebig viele Bytes an den adressierten Slave und empfängt
// anschließend beliebig viele Bytes von dem Slave. 

// slave_address ist die Adresse des Slave in 7bit Schreibweise (0-127).
// send_data zeigt auf die zu sendenden Daten, z.B. ein Array von Bytes. 
// send_bytes gibt an, wieviele Bytes gesendet werden sollen.
// rcv_data zeigt auf einen Puffer, wo die empfangenen Daten abgelegt werden
// sollen, z.B. ein Array von Bytes. 
// rcv_Bytes gibt an, wieviele Bytes empfangen werden sollen.

// Der Rückgabewert zeigt an, wie viele Bytes tatsächlich empfangen wurden.

// Wenn man nur senden will, gibt man rcv_data=0 und rcv_bytes=0 an.
// Wenn man nur empfangen will, gibt man send_data=0 und send_bytes=0 an.
// Es ist erlaubt, bei send_bytes und rcv_bytes Zeiger auf den selben Puffer zu übergeben.

uint8_t i2c_communicate(uint8_t slave_address, void* send_data, uint8_t send_bytes, void* rcv_data, uint8_t rcv_bytes);


#endif // _NIBOBEE_H_
