// Sammlung von mehreren Demo-Programmen.

// Autor: Stefan Frings <stefan@stefanfrings.de>
// Datum: 02.08.2012

// Für folgende Erweiterungen sind nötig:
// Ein Servo von Typ ModelCraft ES-05 an Port X2.2, sowie
// ein Ultraschall-Sensor vom Typ SRF-02 am I2C Port X4.
// Ein Bluetooth Modul BTM-222 oder ein US2Serial Adapterkabel
// am seriellen Port X5.

// Der Servo wird vorne auf die USB Buchse des Roboters geklebt. 
// Er bewegt den Ultraschall-Sensor nach links und rechts.
// Über die serielle Schnittstelle kommuniziert der Roboter mit einem
// PC. Dort also ein Terminal-Programm starten (19200 Baud).

// Nach dem Einschalten drückt man einen Fühler, um eins der Programme zu starten:

// SW1 = Viereck und Kreis fahren
// SW2 = Linie verfolgen
// SW3 = Fernsteuerung via Bluetooth
// SW4 = Herumfahren und Hindernissen ausweichen

// Beim Programm SW1 sendet man mit einem Terminal Programm Buchstaben,
// um den Roboter fern zu steuern:
// f=forward  10cm
// b=backward 10cm
// l=left 15 degrees
// r=right 15 degrees
// q=quit

// Beim Programm SW3 muss der helle Boden mit schwarzem Klebeband markiert
// werden. Der Roboter muss so aufgestellt werden, daß er die Linie wenn
// er maximal 50cm geradeaus fährt. 

// Bei Programm SW4 kann man kann das "Spielfeld" durch eine schwarze 
// Linie auf hellem Boden begrenzen, oder den Roboter auf einem hellen 
// Tisch fahren lassen, dann erkennt er die Tischkanten. 

// Die LED's zeigen folgendes an:

// Beim Start blinken alle vier Led's bei vollen Batterien, bzw. 
// weniger Led's bei schwachen Batterien. Danach:
// LED0 = Mittlerer Lichtsensor (an=hell, aus=dunkel)
// LED1 = Ein Fühler ist betätigt
// LED2 = Motor blockiert
// LED3 = Abstand Ultraschallsensor kleiner als 30cm

// Alle vier Teil-Programme sind mit einer Zustandsmaschine (finite state machine)
// umgesetzt (siehe Wikipedia und http://www.ibweinmann.de/html/echtzeit_auf_muc_.html).

#include "nibobee.h"
#include "fahren.h"
#include "servo.h"
#include <stdio.h>
#include <avr/pgmspace.h>

// Nummer des Servos (die Library unterstützt mehrere)
#define US_SERVO 0

// I2C Adresse des Ultraschall-Sensors (in 7bit Schreibweise)
// Achtung: Im Datenblatt des Sensors ist die Adresse als E0 (in 8bit Schreibweise) angegeben.
#define US_SENSOR 0x70

// Puffer für die I2C Kommunikation (Senden und Empfangen)
uint8_t i2c_buffer[2];

// Ergebnis der letzten Ultraschall-Messung
uint16_t abstand=700;

// Schwellwert für dunkel-Erkennung der drei IR Sensoren
uint16_t dunkel_l=0;
uint16_t dunkel_c=0;
uint16_t dunkel_r=0;


// Aktualisiert alle 100ms die Abstandsmessung via Ultraschall.
// Diese Funktion soll so oft wie möglich immer wieder aufgerufen werden.
void ultraschall_messen() { 
    if ((system_time() % 100) == 0) {
        
        // Lese das Ergebnis der vorherigen Messung aus  
        i2c_buffer[0]=2; // Register, ab dem wir lesen wollen
        uint8_t received=i2c_communicate(US_SENSOR, i2c_buffer, 1, i2c_buffer, 2);
        if (received==2) {
            // Die empfangenen Bytes zu einem 16bit Wert zusammen fügen 
            uint16_t tmp=(i2c_buffer[0]<<8) | i2c_buffer[1];
            // Ungültige Werte ignorieren (nochmal messen)
            if (tmp>10 && tmp<1000) {
                abstand=tmp;
            }
        }      
        
        // Sende Befehl für die nächste Messung
        i2c_buffer[0]=0;  // Register, ab dem wir schreiben wollen
        i2c_buffer[1]=81; // Ping Befehl
        i2c_communicate(US_SENSOR, i2c_buffer, 2, 0, 0);
    }
}


// Zeigt den Status mit den vier LED's an.
// Diese Funktion soll so oft wie möglich immer wieder aufgerufen werden.
void status_anzeigen() {
    set_LED0(analog(LINE_C)<dunkel_c);
    set_LED1(SENS_ANY);
    set_LED2(fahrstatus()==MOTOR_BLOCKIERT || fahrstatus()==STARTET);
    set_LED3(abstand<30);
}


// Kalibriert die Linien-Sensoren, indem 2/3 vom aktuellen Meßwert als
// Schwelle zu dunkel definiert wird.
void kalibriere_linien_sensoren() { 
    dunkel_l=analog(LINE_L)*2/3;
    dunkel_c=analog(LINE_C)*2/3;
    dunkel_r=analog(LINE_R)*2/3;
    printf_P(PSTR("Kalibriert auf: %u, %u, %u\n"),dunkel_l,dunkel_c,dunkel_r); 
}


// Zeigt den Batteriestatus auf der seriellen (Bluetooth) Schnittstelle an
// Diese Funktion blockiert, ist also nicht für die Verwendung innerhalb
// der Zustandsmaschine geeignet.
void batterie_status_bt() {
    set_AREF(REF_256);
    sleep(50);
    printf_P(PSTR("Batterie %u mV\n"),analog(VBAT)*5);
    set_AREF(REF_VCC);
    sleep(50);
}


// Ein Quadrat-Muster abfahren, dann einen Kreis fahren
void prg_viereck() {
    puts_P(PSTR("Viereck-Programm gestartet"));
    sleep(1000);
    
    uint32_t timer=0;
    uint8_t  status=1;
    while (status) {
        switch (status) {
            case  1: // Es wird nach links gedreht
                     vorwaerts(50*CM,SCHNELL);
                     status++;
                     break;
            case  2: // Es wird Vorwärts gefahren
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         rechtskurve(90*GRAD,MITTEL,1000); 
                         status++;
                     }
                     break;
            case  3: // Es wird nach rechts gedreht 
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                        vorwaerts(100*CM,SCHNELL); 
                        status++;
                     }
                     break;
            case  4: // Es wird Vorwärts gefahren
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         rechtskurve(90*GRAD,MITTEL,1000);  
                         status++;
                     }
                     break;
            case  5: // Es wird nach rechts gedreht
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                        vorwaerts(100*CM,SCHNELL); 
                        status++;
                     }
                     break;
            case  6: // Es wird Vorwärts gefahren
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         rechtskurve(90*GRAD,MITTEL,1000);   
                         status++;
                     }
                     break;
            case  7: // Es wird nach rechts gedreht
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                        vorwaerts(100*CM,SCHNELL); 
                        status++;
                     }
                     break;
            case  8: // Es wird Vorwärts gefahren
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         rechtskurve(90*GRAD,MITTEL,1000);  
                         status++;
                     }
                     break;
            case  9: // Es wird nach rechts gedreht
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         vorwaerts(50*CM,SCHNELL);
                         status++;
                     }
                     break;
            case 10: // Es wird Vorwärts gefahren
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         // Warte zwei Sekunden
                         timer=system_time()+1000; 
                         status++;
                     }
                     break;
            case 11: // Timer wird abgewartet
                     if (system_time()>=timer) {
                         rechtskurve(360*GRAD,MITTEL,600);
                         status++;
                     }
                     break; 
            case 12: // Rechtskurve wird gefahren 
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         status++;
                     }
                     break; 
            default: // Ende
                     status=0;
                     
        }
        
        // Die folgenden Funktionen werden unabhängig vom Status ständig wiederholt aufgerufen
        fahren_regeln();
        ultraschall_messen();
        status_anzeigen();
    }    
}

// Linie auf dem Boden suchen und verfolgen
void prg_linie() {
    puts_P(PSTR("Linien-Verfolge-Programm gestartet"));
    sleep(1000);
    kalibriere_linien_sensoren();
    
    uint8_t  status=1;
    while (status) {
        switch (status) {            
            case  1: // Start
                     // Suche die Linie vorne in maximal 50cm Entfernung
                     vorwaerts(50*CM,LANGSAM);
                     status++;
                     break;
            case  2: // Linie wird gesucht
                     // Wenn Linie gefunden, dann vorwärts fahren
                     if (analog(LINE_C)<dunkel_c) {
                         set_DIR_R(0);
                         set_DIR_L(1);
                         status++;
                     }
                     // Wenn Linie nicht gefunden, dann State-Machine beenden 
                     else if (fahrstatus()>=ZIEL_ERREICHT) {
                         status=0;
                     } 
                     break;
            case  3: // Es wird gefahren
                     // Wenn die Linie nicht mehr in der Mitte ist, wo ist sie dann?
                     if (analog(LINE_C)>dunkel_c) {
                         if (analog(LINE_L)>dunkel_l && analog(LINE_R)<dunkel_r) {
                             // Linie ist rechts, nach rechts lenken
                             PWM_L=1000;
                             PWM_R=400;
                         }
                         else if (analog(LINE_L)<dunkel_l && analog(LINE_R)>dunkel_r) {
                             // Linie ist links, nach links lenken
                             PWM_L=400;
                             PWM_R=1000;
                         }
                         // Wenn die Linie seit mehr als 30cm verloren ist, beende die Fahrt
                         else if ((odometer_left()>30*CM) || (odometer_right()>30*CM)) {
                             PWM_L=0;
                             PWM_R=0;
                             status++;
                         }
                     }
                     else {
                        // Wir sind auf der Linie, geradeaus fahren
                        reset_odometer();
                        PWM_L=1000;
                        PWM_R=1000;
                     }
                     break;
            default: // Ende
                     status=0;
        }
        
        // Die folgenden Funktionen werden unabhängig vom Status ständig wiederholt aufgerufen
        fahren_regeln();
        ultraschall_messen();
        status_anzeigen();
        // Wenn irgendwo angestoßen, dann abbrechen
        if (SENS_ANY) {
            stop();
            status=0;
        }
    }
}

// Fernsteuerung über Bluetooth (ganz rudimentär)
void prg_bluetooth() {
    puts_P(PSTR("Bluetooth-Programm gestartet"));
    puts_P(PSTR("f=forward  10cm"));
    puts_P(PSTR("b=backward 10cm"));
    puts_P(PSTR("l=left 15 degrees"));
    puts_P(PSTR("r=right 15 degrees"));
    puts_P(PSTR("q=quit"));
    sleep(1000);
    
    uint8_t  status=1;
    while (status) {
        switch (status) {            
            case  1: // Start
                     // Wartet auf dem Empfang eines Befehls
                     if (RECEIVED) {
                         // Führe den empfangenen Befehl aus
                         char c=getchar();                         
                         switch (c) {
                            case 'f': vorwaerts(10*CM,MITTEL); 
                                      status++;
                                      break;
                            case 'b': rueckwaerts(10*CM,MITTEL); 
                                      status++;
                                      break;
                            case 'l': rotiere(-15*GRAD,MITTEL); 
                                      status++;
                                      break;
                            case 'r': rotiere(15*GRAD,MITTEL); 
                                      status++;
                                      break;
                            case 'q': status=0; //Beenden
                        } 
                     }
                     break;
            case  2: // Es wird gefahren
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         // Nochmal auf Befehl warten
                         status=1;
                     }
                     break;
            default: // Ende
                     status=0;
        }
        
        // Die folgenden Funktionen werden unabhängig vom Status ständig wiederholt aufgerufen
        fahren_regeln();
        ultraschall_messen();
        status_anzeigen();
    }
}


// Herumfahren und Hindernissen ausweichen
void prg_ausweichen() {
    puts_P(PSTR("Ausweichen-Programm gestartet"));
    sleep(1000);
    kalibriere_linien_sensoren();
    
    uint8_t status=1;
    uint8_t ursache=0;
    uint32_t timer=0;
    uint16_t abstand_l=0;
    uint16_t abstand_r=0;
    while (status) {
        switch (status) {            
            case  1: // Start
                     vorwaerts(500*CM,SUPERSCHNELL);
                     status++;                     
                     break;
            case  2: // Fährt vorwärts
                     // Wenn Hindernis getroffen, merke die Ursache
                     if (SENS_LEFT)
                         ursache=1;
                     else if (SENS_RIGHT)
                         ursache=2;
                     else if (abstand<35)
                         ursache=3;
                     else if (analog(LINE_C)<dunkel_c)
                         ursache=4;
                     else if (fahrstatus()==MOTOR_BLOCKIERT)
                         ursache=5;
                     else 
                         ursache=0; // kein Hindernis
                     // Wenn Hindernis getroffen, dann setze zurück
                     if (ursache>0) {
                         rueckwaerts(20*CM,MITTEL);
                         status++;
                     }  
                     // Wenn Ziel ohne Hindernis erreicht, weiter fahren
                     else if (fahrstatus()>=ZIEL_ERREICHT) {
                         status=1;
                     }
                     break;
            case  3: // Fährt rückwärts
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         switch (ursache) {
                             case 1: // Links angestoßen
                                     rotiere(45*GRAD,MITTEL);
                                     status++;
                                     break;
                             case 2: // Rechts angestoßen
                                     rotiere(-45*GRAD,MITTEL);
                                     status++;
                                     break;
                             case 3: // Ultraschall Hindernis erkannt
                                     status=5;                                
                                     break;
                             case 4: // Dunkle Linie erkannt
                                     rotiere(180*GRAD,MITTEL);
                                     status++;
                                     break;
                             case 5: // Motor klemmt
                                     rotiere(180*GRAD,MITTEL);
                                     status++;
                                     break;
                         }
                     }
                     break;
            case  4: // Dreht sich weg
                     // Warte bis fertig gedreht, fahre dann weiter
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         status=1;
                     }
                     break;
            case  5: // Ultraschall Hindernis erkannt, suche einen Ausweg
                     puts_P(PSTR("Ultraschall Hindernis"));
                     // Schau nach links
                     set_servo(US_SERVO,-90);
                     timer=system_time()+400;
                     status++;
                     break;
            case  6: // Servo dreht sich nach links
                     if (system_time()>=timer) {
                         abstand_l=abstand;
                         printf_P(PSTR("Abstand links: %u cm\n"),abstand_l);
                         // Schau nach rechts
                         set_servo(US_SERVO,90);
                         timer=system_time()+800;
                         status++;
                     }
                     break;
            case  7: // Servo dreht sich nach rechts
                     if (system_time()>=timer) {
                         abstand_r=abstand;
                         printf_P(PSTR("Abstand rechts: %u cm\n"),abstand_r);
                         // Servo wieder nach vorne richten
                         set_servo(US_SERVO,0);
                         // Dorthin drehen, wo mehr Platz ist
                         if (abstand_l>abstand_r && abstand_l>35) {
                             rotiere(-90*GRAD,MITTEL);
                         }
                         else if (abstand_r>35) {
                             rotiere(90*GRAD,MITTEL);
                         }
                         else {
                             // links und rechts ist wenig Platz, umkehren
                             rotiere(180*GRAD,MITTEL);
                         }
                         status++;
                     }
                     break;
            case  8: // Fahrzeug dreht sich dorthin, wo mehr Platz ist                     
                     if (fahrstatus()>=ZIEL_ERREICHT) {
                         status=1;
                     }
                     break;
            default: // Ende
                     status=0;
        }
        
        // Die folgenden Funktionen werden unabhängig vom Status ständig wiederholt aufgerufen
        fahren_regeln();
        ultraschall_messen();
        status_anzeigen();
    }
}


int main() {
    puts_P(PSTR("reset"));
    while (1) {
        batterie_status();
        batterie_status_bt(); 
        set_servo(US_SERVO,0);
        kalibriere_linien_sensoren();
        
        // Warte auf Programm-Auswahl
        puts_P(PSTR("Waehle ein Programm mit den Fuehlern"));
        puts_P(PSTR("SW1 = Viereck fahren, dann 90 Grad Linkskurve"));
        puts_P(PSTR("SW2 = Linie verfolgen"));
        puts_P(PSTR("SW3 = Fernsteuerung via Bluetooth"));
        puts_P(PSTR("SW4 = Herumfahren und Hindernissen ausweichen"));
        
        // Warte auf Betätigung eines Fühlers
        do {
            ultraschall_messen();
            status_anzeigen();
        } 
        while (!SENS_ANY);
        if (SENS_SW1) prg_viereck(); 
        else if (SENS_SW2) prg_linie(); 
        else if (SENS_SW3) prg_bluetooth();
        else if (SENS_SW4) prg_ausweichen(); 
    }
}
